/*
  This file is part of TALER
  Copyright (C) 2014-2026 Taler Systems SA

  TALER is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
  Foundation; either version 3, or (at your option) any later version.

  TALER is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  TALER; see the file COPYING.  If not, see
  <http://www.gnu.org/licenses/>
*/
/**
 * @file lib/exchange_api_get-reserves-RESERVE_PUB.c
 * @brief Implementation of the GET /reserves/$RESERVE_PUB requests
 * @author Christian Grothoff
 */
#include <jansson.h>
#include <microhttpd.h> /* just for HTTP status codes */
#include <gnunet/gnunet_util_lib.h>
#include <gnunet/gnunet_json_lib.h>
#include <gnunet/gnunet_curl_lib.h>
#include "taler/taler_json_lib.h"
#include "exchange_api_handle.h"
#include "taler/taler_signatures.h"
#include "exchange_api_curl_defaults.h"


/**
 * @brief A GET /reserves/$RESERVE_PUB Handle
 */
struct TALER_EXCHANGE_GetReservesHandle
{

  /**
   * Base URL of the exchange.
   */
  char *base_url;

  /**
   * The url for this request.
   */
  char *url;

  /**
   * CURL context to use.
   */
  struct GNUNET_CURL_Context *ctx;

  /**
   * Handle for the request.
   */
  struct GNUNET_CURL_Job *job;

  /**
   * Function to call with the result.
   */
  TALER_EXCHANGE_GetReservesCallback cb;

  /**
   * Closure for @e cb.
   */
  TALER_EXCHANGE_GET_RESERVES_RESULT_CLOSURE *cb_cls;

  /**
   * Public key of the reserve we are querying.
   */
  struct TALER_ReservePublicKeyP reserve_pub;

  /**
   * Options for the request.
   */
  struct
  {
    /**
     * How long to wait for an answer (enables long polling).
     */
    struct GNUNET_TIME_Relative timeout;
  } options;

};


/**
 * We received an #MHD_HTTP_OK status code. Handle the JSON response.
 *
 * @param grh handle of the request
 * @param j JSON response
 * @return #GNUNET_OK on success
 */
static enum GNUNET_GenericReturnValue
handle_reserves_get_ok (struct TALER_EXCHANGE_GetReservesHandle *grh,
                        const json_t *j)
{
  struct TALER_EXCHANGE_GetReservesResponse rs = {
    .hr.reply = j,
    .hr.http_status = MHD_HTTP_OK
  };
  struct GNUNET_JSON_Specification spec[] = {
    TALER_JSON_spec_amount_any ("balance",
                                &rs.details.ok.balance),
    GNUNET_JSON_spec_mark_optional (
      GNUNET_JSON_spec_string (
        "last_origin",
        (const char **) &rs.details.ok.last_origin.full_payto),
      NULL),
    GNUNET_JSON_spec_end ()
  };

  if (GNUNET_OK !=
      GNUNET_JSON_parse (j,
                         spec,
                         NULL,
                         NULL))
  {
    GNUNET_break_op (0);
    return GNUNET_SYSERR;
  }
  grh->cb (grh->cb_cls,
           &rs);
  grh->cb = NULL;
  return GNUNET_OK;
}


/**
 * Function called when we're done processing the
 * HTTP GET /reserves/$RESERVE_PUB request.
 *
 * @param cls the `struct TALER_EXCHANGE_GetReservesHandle`
 * @param response_code HTTP response code, 0 on error
 * @param response parsed JSON result, NULL on error
 */
static void
handle_reserves_get_finished (void *cls,
                              long response_code,
                              const void *response)
{
  struct TALER_EXCHANGE_GetReservesHandle *grh = cls;
  const json_t *j = response;
  struct TALER_EXCHANGE_GetReservesResponse rs = {
    .hr.reply = j,
    .hr.http_status = (unsigned int) response_code
  };

  grh->job = NULL;
  switch (response_code)
  {
  case 0:
    rs.hr.ec = TALER_EC_GENERIC_INVALID_RESPONSE;
    break;
  case MHD_HTTP_OK:
    if (GNUNET_OK !=
        handle_reserves_get_ok (grh,
                                j))
    {
      rs.hr.http_status = 0;
      rs.hr.ec = TALER_EC_GENERIC_REPLY_MALFORMED;
    }
    break;
  case MHD_HTTP_BAD_REQUEST:
    /* This should never happen, either us or the exchange is buggy
       (or API version conflict); just pass JSON reply to the application */
    rs.hr.ec = TALER_JSON_get_error_code (j);
    rs.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  case MHD_HTTP_NOT_FOUND:
    /* Nothing really to verify, this should never
       happen, we should pass the JSON reply to the application */
    rs.hr.ec = TALER_JSON_get_error_code (j);
    rs.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  case MHD_HTTP_INTERNAL_SERVER_ERROR:
    /* Server had an internal issue; we should retry, but this API
       leaves this to the application */
    rs.hr.ec = TALER_JSON_get_error_code (j);
    rs.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  default:
    /* unexpected response code */
    GNUNET_break_op (0);
    rs.hr.ec = TALER_JSON_get_error_code (j);
    rs.hr.hint = TALER_JSON_get_error_hint (j);
    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                "Unexpected response code %u/%d for GET %s\n",
                (unsigned int) response_code,
                (int) rs.hr.ec,
                grh->url);
    break;
  }
  if (NULL != grh->cb)
  {
    grh->cb (grh->cb_cls,
             &rs);
    grh->cb = NULL;
  }
  TALER_EXCHANGE_get_reserves_cancel (grh);
}


struct TALER_EXCHANGE_GetReservesHandle *
TALER_EXCHANGE_get_reserves_create (
  struct GNUNET_CURL_Context *ctx,
  const char *url,
  const struct TALER_ReservePublicKeyP *reserve_pub)
{
  struct TALER_EXCHANGE_GetReservesHandle *grh;

  grh = GNUNET_new (struct TALER_EXCHANGE_GetReservesHandle);
  grh->ctx = ctx;
  grh->base_url = GNUNET_strdup (url);
  grh->reserve_pub = *reserve_pub;
  return grh;
}


enum GNUNET_GenericReturnValue
TALER_EXCHANGE_get_reserves_set_options_ (
  struct TALER_EXCHANGE_GetReservesHandle *grh,
  unsigned int num_options,
  const struct TALER_EXCHANGE_GetReservesOptionValue *options)
{
  for (unsigned int i = 0; i < num_options; i++)
  {
    switch (options[i].option)
    {
    case TALER_EXCHANGE_GET_RESERVES_OPTION_END:
      return GNUNET_OK;
    case TALER_EXCHANGE_GET_RESERVES_OPTION_TIMEOUT:
      grh->options.timeout = options[i].details.timeout;
      break;
    default:
      GNUNET_break (0);
      return GNUNET_SYSERR;
    }
  }
  return GNUNET_OK;
}


enum TALER_ErrorCode
TALER_EXCHANGE_get_reserves_start (
  struct TALER_EXCHANGE_GetReservesHandle *grh,
  TALER_EXCHANGE_GetReservesCallback cb,
  TALER_EXCHANGE_GET_RESERVES_RESULT_CLOSURE *cb_cls)
{
  char arg_str[sizeof (struct TALER_ReservePublicKeyP) * 2 + 16];
  CURL *eh;
  unsigned int tms
    = (unsigned int) grh->options.timeout.rel_value_us
      / GNUNET_TIME_UNIT_MILLISECONDS.rel_value_us;

  if (NULL != grh->job)
  {
    GNUNET_break (0);
    return TALER_EC_GENERIC_INTERNAL_INVARIANT_FAILURE;
  }
  grh->cb = cb;
  grh->cb_cls = cb_cls;
  {
    char pub_str[sizeof (struct TALER_ReservePublicKeyP) * 2];
    char *end;
    char timeout_str[32];

    end = GNUNET_STRINGS_data_to_string (
      &grh->reserve_pub,
      sizeof (grh->reserve_pub),
      pub_str,
      sizeof (pub_str));
    *end = '\0';
    GNUNET_snprintf (arg_str,
                     sizeof (arg_str),
                     "reserves/%s",
                     pub_str);
    GNUNET_snprintf (timeout_str,
                     sizeof (timeout_str),
                     "%u",
                     tms);
    grh->url = TALER_url_join (grh->base_url,
                               arg_str,
                               "timeout_ms",
                               (0 == tms)
                                ? NULL
                                : timeout_str,
                               NULL);
  }
  if (NULL == grh->url)
    return TALER_EC_GENERIC_CONFIGURATION_INVALID;
  eh = TALER_EXCHANGE_curl_easy_get_ (grh->url);
  if (NULL == eh)
    return TALER_EC_GENERIC_CONFIGURATION_INVALID;
  if (0 != tms)
  {
    GNUNET_break (CURLE_OK ==
                  curl_easy_setopt (eh,
                                    CURLOPT_TIMEOUT_MS,
                                    (long) (tms + 100L)));
  }
  grh->job = GNUNET_CURL_job_add (grh->ctx,
                                  eh,
                                  &handle_reserves_get_finished,
                                  grh);
  if (NULL == grh->job)
    return TALER_EC_GENERIC_INTERNAL_INVARIANT_FAILURE;
  return TALER_EC_NONE;
}


void
TALER_EXCHANGE_get_reserves_cancel (
  struct TALER_EXCHANGE_GetReservesHandle *grh)
{
  if (NULL != grh->job)
  {
    GNUNET_CURL_job_cancel (grh->job);
    grh->job = NULL;
  }
  GNUNET_free (grh->url);
  GNUNET_free (grh->base_url);
  GNUNET_free (grh);
}


/* end of exchange_api_get-reserves-RESERVE_PUB.c */
